import numpy as np
import pytest

from cosmic_toolbox.arraytools import (
    add_cols,
    arr2rec,
    arr_to_rec,
    delete_cols,
    dict2class,
    dict2rec,
    rec2arr,
    rec2dict,
)


@pytest.fixture
def example_array():
    return np.array([[1, 2], [3, 4], [5, 6]])


@pytest.fixture
def example_names():
    return ["col1", "col2"]


@pytest.fixture
def example_dict():
    return {"col1": [1, 2, 3], "col2": [4, 5, 6]}


@pytest.fixture
def example_dtype():
    return np.dtype([("col1", int), ("col2", int)])


def compare_dict(dict1, dict2):
    return np.all([dict1[key] == dict2[key] for key in dict1.keys() if key in dict2])


def test_arr2rec(example_array, example_names):
    expected = np.rec.array([(1, 2), (3, 4), (5, 6)], names=example_names)
    assert np.array_equal(arr2rec(example_array, example_names), expected)


def test_rec2arr(example_array, example_names):
    rec_array = np.rec.array([(1, 2), (3, 4), (5, 6)], names=example_names)
    expected = np.array([[1, 2], [3, 4], [5, 6]])
    assert np.array_equal(rec2arr(rec_array), expected)


def test_dict2rec(example_dict):
    expected = np.rec.array([(1, 4), (2, 5), (3, 6)], names=["col1", "col2"])
    assert np.array_equal(dict2rec(example_dict), expected)


def test_dict2rec_extended():
    x = {"a": np.zeros(5), "b": 3}
    x = dict2rec(x)
    assert np.all(x["a"] == np.zeros(5))
    assert np.all(x["b"] == np.array([3, 3, 3, 3, 3]))

    x = {"a": np.zeros(5), "b": np.ones((5, 2))}
    x = dict2rec(x)
    assert np.all(x["a"] == np.zeros(5))
    assert np.all(x["b"] == np.ones((5, 2)))


def test_rec2dict(example_dtype):
    rec_array = np.rec.array([(1, 4), (2, 5), (3, 6)], dtype=example_dtype)
    expected = {"col1": np.array([1, 2, 3]), "col2": np.array([4, 5, 6])}
    predicted = rec2dict(rec_array)
    assert compare_dict(expected, predicted)


def test_dict2class():
    # Test case 1
    d1 = {"a": [1, 2, 3], "b": 4}
    c1 = dict2class(d1)
    assert c1.a == [1, 2, 3]
    assert c1.b == 4

    # Test case 2
    d2 = {"x": "hello", "y": {"a": 5, "b": 6, "c": 7}}
    c2 = dict2class(d2)
    assert c2.x == "hello"
    assert c2.y == {"a": 5, "b": 6, "c": 7}

    # Test case 3: Empty dictionary
    d3 = {}
    c3 = dict2class(d3)
    assert hasattr(c3, "a") is False
    assert hasattr(c3, "b") is False


@pytest.fixture
def sample_data():
    data = np.array(
        [(1, 2, 3), (4, 5, 6), (7, 8, 9)], dtype=[("A", int), ("B", int), ("C", int)]
    )
    return data


# Define the tests using the fixture
def test_delete_cols_single_column(sample_data):
    result = delete_cols(sample_data, ["B"])
    expected_result = np.array([(1, 3), (4, 6), (7, 9)], dtype=[("A", int), ("C", int)])
    assert np.array_equal(result, expected_result)


def test_delete_cols_multiple_columns(sample_data):
    result = delete_cols(sample_data, ["B", "C"])
    expected_result = np.array([(1,), (4,), (7,)], dtype=[("A", int)])
    assert np.array_equal(result, expected_result)


def test_delete_cols_non_existent_column(sample_data):
    result = delete_cols(sample_data, ["D"])
    assert np.array_equal(result, sample_data)  # Should not modify the original data


def test_delete_cols_all_columns(sample_data):
    result = delete_cols(sample_data, ["A", "B", "C"])
    expected_result = np.array([], dtype=[])
    assert np.array_equal(result, expected_result)


def test_add_cols_single_scalar(sample_data):
    # Test case 1: Adding a single column with scalar data
    result = add_cols(sample_data, ["D"], data=0)
    assert "D" in result.dtype.names
    assert np.array_equal(result["D"], np.zeros_like(result["A"]))


def test_add_cols_multiple_scalars(sample_data):
    # Test case 2: Adding multiple columns with scalar data
    result = add_cols(sample_data, ["D", "E"], data=1)
    assert "D" in result.dtype.names
    assert "E" in result.dtype.names
    assert np.array_equal(result["D"], np.ones_like(result["A"]))
    assert np.array_equal(result["E"], np.ones_like(result["A"]))


def test_add_cols_single_1d_array(sample_data):
    # Test case 3: Adding a single column with 1D array data
    data_array = np.array([10, 20, 30])
    result = add_cols(sample_data, ["F"], data=data_array)
    assert "F" in result.dtype.names
    assert np.array_equal(result["F"], data_array)


def test_add_cols_multiple_2d_array(sample_data):
    # Test case 4: Adding multiple columns with 2D array data
    data_array = np.array([[10, 40, 70], [20, 50, 80]])
    result = add_cols(sample_data, ["F", "G"], data=data_array)
    assert "F" in result.dtype.names
    assert "G" in result.dtype.names
    assert np.array_equal(result["F"], data_array[0])
    assert np.array_equal(result["G"], data_array[1])


def test_arr_to_rec():
    arr = np.array([[1, 2], [3, 4]])
    dtype = [("a", np.int64), ("b", np.int64)]
    newrecarray = arr_to_rec(arr, dtype)
    assert np.all(newrecarray["a"] == arr[:, 0])
    assert np.all(newrecarray["b"] == arr[:, 1])



"""
author: Tomasz Kacprzak
"""

import sys
import warnings

import numpy as np
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator, Rbf
from sklearn.ensemble import RandomForestRegressor
from sklearn.neighbors import BallTree, KNeighborsRegressor, RadiusNeighborsRegressor
from sklearn.preprocessing import MinMaxScaler

from cosmic_toolbox.logger import get_logger

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("once", category=UserWarning)
LOGGER = get_logger(__file__)


class Rbft:
    def __init__(self, points, values, **kw_rbf):
        self.points = points
        self.values = values
        self.bounds = [np.min(points, axis=0), np.max(points, axis=0)]
        self.scaler = MinMaxScaler()
        self.scaler.fit(self.points)
        self.points = self.scaler.transform(self.points)
        self.interp = Rbf(*list(self.points.T), self.values, **kw_rbf)

    def __call__(self, points, **kw_rbf):
        values_pred = np.zeros(len(points))
        select = self._in_bounds(points)
        values_pred[~select] = -np.inf

        points = self.scaler.transform(points)
        values_pred[select] = self.interp(*list(points[select].T))
        return values_pred

    def _in_bounds(self, x):
        return np.all(x > self.bounds[0], axis=1) & np.all(x < self.bounds[1], axis=1)


def query_split(X, tree, k, n_proc):
    nx = X.shape[0]
    n_per_batch = int(np.ceil(nx / n_proc))
    LOGGER.info(
        "querying BallTree with a pool for n_grid={} "
        "n_proc={} n_per_batch={} n_neighbors={}".format(nx, n_proc, n_per_batch, k)
    )
    X_chunks = [X[(i * n_per_batch) : (i + 1) * n_per_batch, :] for i in range(n_proc)]
    from functools import partial
    from multiprocessing import Pool

    f = partial(query_batch, tree=tree, k=k, n_per_batch=100000)
    with Pool(n_proc) as pool:
        list_y = pool.map(f, X_chunks)

    distances = np.concatenate([list_y[i][0] for i in range(n_proc)])
    indices = np.concatenate([list_y[i][1] for i in range(n_proc)])
    return distances, indices


def query_batch(X, tree, k=100, n_per_batch=10000):
    nx = X.shape[0]
    n_batches = int(np.ceil(nx / n_per_batch))
    indices = np.zeros([nx, k], dtype=np.int)
    distances = np.zeros([nx, k])
    for i in range(n_batches):
        si, ei = i * n_per_batch, (i + 1) * n_per_batch
        Xq = X[si:ei, :]
        if len(Xq) > 0:
            dist, ind = tree.query(Xq, k=k)
            indices[si:ei, :] = ind
            distances[si:ei, :] = dist
            LOGGER.info("batch={:>6}/{:>6}".format(i, n_batches))

    return distances, indices


def predict_with_neighbours(y, ind, dist):
    yn = y[ind]
    wn = 1.0 / dist
    wn[wn == 0] = 1e10
    yi = np.average(yn, weights=wn, axis=1)

    return yi


def predict_knn_balltree(Xi, X, y, n_neighbors, tree):
    nx, nd = Xi.shape
    dist, ind = tree.query(Xi, k=n_neighbors)
    yi = predict_with_neighbours(y, ind, dist)
    return yi


def predict_knn_linear(Xi, X, y, n_neighbors, tree):
    nx, nd = Xi.shape
    dist, ind = tree.query(Xi, k=n_neighbors)
    X_nearest = X[ind, :]
    y_nearest = y[ind]
    yi = np.full(nx, -np.inf)
    n_nan = 0

    for i in range(nx):
        Xn = X_nearest[i, :]
        yn = y_nearest[i, :]
        interp = LinearNDInterpolator(Xn, yn)
        yi[i] = interp(Xi[i, :])
        if ~np.isfinite(yi[i]):
            n_nan += 1
        sys.stdout.write("\r{}/{} {}".format(i, nx, n_nan))

    return yi


class MultiInterp:
    def __init__(self, X, y, method="nn", **kw):
        self.X = X.copy()
        self.y = y.copy()
        self.bounds = [np.min(X, axis=0), np.max(X, axis=0)]
        self.scaler = MinMaxScaler()
        self.scaler.fit(self.X)
        self.X = self.scaler.transform(self.X)
        self.interp = None
        self.method = method
        self.kw = kw
        self.init_interp(**kw)

    def __call__(self, Xi, **kw):
        yi = np.zeros(len(Xi))
        select = self._in_bounds(Xi)
        yi[~select] = -np.inf

        Xi = self.scaler.transform(Xi.copy())

        if self.method.lower() == "rbf":
            yi[select] = self.interp(*list(Xi[select, :].T), **kw)

        elif self.method.lower() == "rbft":
            yi[select] = self.interp(Xi[select, :], **kw)

        elif self.method.lower() in ["knn_regression", "rnn_regression"]:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                yi[select] = self.interp.predict(Xi[select, :], **kw)

        elif self.method.lower() in ["nn", "linear"]:
            yi[select] = self.interp(Xi[select, :], **kw)

        elif self.method.lower() == "knn_linear":
            yi[select] = predict_knn_linear(
                Xi[select, :],
                self.X,
                self.y,
                self.n_neighbors,
                self.interp,
                **kw,
            )

        elif self.method.lower() == "random_forest":
            yi[select] = self.interp.predict(Xi[select, :], **kw)

        elif self.method.lower() == "knn_balltree":
            yi[select] = predict_knn_balltree(
                Xi[select, :],
                self.X,
                self.y,
                self.n_neighbors,
                self.interp,
                **kw,
            )

        else:
            raise Exception(f"unknown interp method {self.method}")

        return yi

    def _in_bounds(self, X):
        return np.all(X > self.bounds[0], axis=1) & np.all(X < self.bounds[1], axis=1)

    def init_interp(self, **kw):
        if self.method.lower() == "rbf":
            self.interp = Rbf(*list(self.X.T), self.y, **kw)

        elif self.method.lower() == "rbft":
            self.interp = Rbft(points=self.X, values=self.y, **kw)

        elif self.method.lower() == "knn":
            self.interp = KNeighborsRegressor(**kw)
            self.interp.fit(self.X, self.y)

        elif self.method.lower() == "knn_regression":
            self.interp = KNeighborsRegressor(**kw)
            self.interp.fit(self.X, self.y)

        elif self.method.lower() == "rnn_regression":
            kw.setdefault("radius", 0.1)
            self.interp = RadiusNeighborsRegressor(**kw)
            self.interp.fit(self.X, self.y)

        elif self.method.lower() == "nn":
            self.interp = NearestNDInterpolator(self.X, self.y, **kw)

        elif self.method.lower() == "linear":
            self.interp = LinearNDInterpolator(self.X, self.y, **kw)

        elif self.method.lower() == "knn_linear":
            # self.slice_linear_upsampling(n_repeat=1)
            n_neighbors = kw.pop("n_neighbors", self.X.shape[1] * 3)
            self.interp = BallTree(self.X, **kw)
            self.n_neighbors = n_neighbors

        elif self.method.lower() == "random_forest":
            self.interp = RandomForestRegressor(**kw)
            self.interp.fit(self.X, self.y)

        elif self.method.lower() == "knn_balltree":
            n_neighbors = kw.pop("n_neighbors", self.X.shape[1] * 3)
            self.interp = BallTree(self.X, **kw)
            self.n_neighbors = n_neighbors

        else:
            raise Exception(f"unknown interp method {self.method}")

    def precompute_grid_neighbors(self, Xn, n_neighbors=100, n_proc=1):
        assert self.method == "knn_balltree"

        Xn = self.scaler.transform(Xn.copy())
        dist, ind = query_split(tree=self.interp, X=Xn, k=n_neighbors, n_proc=n_proc)

        self.neighbors_ind = ind
        self.neighbors_dist = dist.astype(np.float32)
        self.neighbors_Xn = Xn.astype(np.float32)

    def interpolate_grid_neighbours(self, y, n_neighbors=None):
        assert len(y) == len(self.X)
        if n_neighbors is None:
            n_neighbors = self.n_neighbors

        nn = self.neighbors_ind.shape[1]
        if n_neighbors > nn:
            raise Exception("number of available neighbors {}".format(nn))
        yi = predict_with_neighbours(
            y,
            self.neighbors_ind[:, :n_neighbors],
            self.neighbors_dist[:, :n_neighbors],
        )

        return yi

    def slice_linear_upsampling(self, n_repeat=1, n_neighbors=1):
        for i in range(n_repeat):
            bt = BallTree(self.X)
            dist, ids = bt.query(self.X, k=n_neighbors + 1)
            list_Xn = [self.X]
            list_yn = [self.y]
            for j in range(1, n_neighbors + 1):
                Xn = (self.X + self.X[ids[:, j], :]) / 2.0
                yn = (self.y + self.y[ids[:, j]]) / 2.0
                list_Xn += [Xn]
                list_yn += [yn]

                # Xp = self.X[ids,:]
                # yp = self.y[ids]
                # Xn = Xp.mean(axis=1)
                # yn = yp.mean(axis=1)
            # Xc = np.concatenate([self.X, Xn], axis=0)
            # yc = np.concatenate([self.y, yn])
            Xn = np.concatenate(list_Xn, axis=0)
            yn = np.concatenate(list_yn, axis=0)
            self.X = Xn
            self.y = yn
            print(Xn.shape, yn.shape)



import numpy as np
from scipy.stats import norm
from sklearn.mixture import GaussianMixture


def scale_fwd(x, param_bounds, param_bounds_trans):
    xs = x - param_bounds[0]
    xs = (
        xs
        / (param_bounds[1] - param_bounds[0])
        * (param_bounds_trans[1] - param_bounds_trans[0])
    )
    xs = xs + param_bounds_trans[0]
    return xs


def scale_inv(x, param_bounds, param_bounds_trans):
    xs = x - param_bounds_trans[0]
    xs = (
        xs
        * (param_bounds[1] - param_bounds[0])
        / (param_bounds_trans[1] - param_bounds_trans[0])
    )
    xs = xs + param_bounds[0]
    return xs


def trans_fwd(x, param_bounds, param_bounds_trans):
    xs = scale_fwd(x, param_bounds, param_bounds_trans)
    ppfx = norm.ppf(xs)
    if np.any(~np.isfinite(ppfx)):
        import ipdb

        ipdb.set_trace()
    return ppfx


def trans_inv(x, param_bounds, param_bounds_trans):
    xi = norm.cdf(x)
    xi = scale_inv(xi, param_bounds, param_bounds_trans)
    return xi


class TransformedGaussianMixture:
    def __init__(self, param_bounds=None, *args, **kwargs):
        self.eps = 1e-8
        self.bounds_trans = [self.eps, 1 - self.eps]
        self.param_bounds = param_bounds

        self.gm = GaussianMixture(*args, **kwargs)

    def set_bounds(self, X):
        if self.param_bounds is None:
            self.param_bounds = np.array(
                [
                    np.min(X, axis=0) - 10 * self.eps,
                    np.max(X, axis=0) + 10 * self.eps,
                ]
            ).T

    def fit(self, X, y=None):
        self.set_bounds(X)
        X_trans = self._transform_params_forward(X)
        return self.gm.fit(X_trans, y)

    def fit_predict(self, X, y=None):
        self.set_bounds(X)
        X_trans = self._transform_params_forward(X)
        return self.gm.fit_predict(X_trans, y)

    def predict_proba(self, X):
        X_trans = self._transform_params_forward(X)
        return self.gm.predict_proba(X_trans)

    def sample(self, n_samples=1):
        # print('ol')
        # import chaospy
        # means = self.gm.means_[:, :2]
        # covariances = self.gm.covariances_[:, :2, :2]
        # self.chaos_gm = chaospy.GaussianMixture(means, covariances)
        # X_trans, y = self.chaos_gm.sample(n_samples, rule="halton")
        X_trans, y = self.gm.sample(n_samples)
        X = self._transform_params_inverse(X_trans)
        return X, y

    def score(self, X, y=None):
        X_trans = self._transform_params_forward(X)
        return self.gm.score(X_trans)

    def score_samples(self, X):
        X_trans = self._transform_params_forward(X)
        return self.gm.score_samples(X_trans)

    def set_params(self, **params):
        return self.gm.set_params(**params)

    def get_params(self, deep=True):
        return self.gm.get_params(deep)

    def bic(self, X):
        X_trans = self._transform_params_forward(X)
        return self.gm.bic(X_trans)

    def aic(self, X):
        X_trans = self._transform_params_forward(X)
        return self.gm.aic(X_trans)

    def _transform_params_forward(self, X):
        X_trans = X.copy()
        for i in range(X.shape[1]):
            X_trans[:, i] = trans_fwd(
                x=np.array(X[:, i]),
                param_bounds=self.param_bounds[i],
                param_bounds_trans=self.bounds_trans,
            )

        return X_trans

    def _transform_params_inverse(self, X_trans):
        X = X_trans.copy()
        for i in range(X.shape[1]):
            X[:, i] = trans_inv(
                x=np.array(X_trans[:, i]),
                param_bounds=self.param_bounds[i],
                param_bounds_trans=self.bounds_trans,
            )

        return X



# Copyright (C) 2023 ETH Zurich
# Institute for Particle Physics and Astrophysics
# Author: Silvan Fischbacher

import argparse
import ast
import re
from multiprocessing import Pool

from cosmic_toolbox.logger import get_logger

LOGGER = get_logger(__file__)


def arg_str_to_dict(arg_str):
    """
    Converts a string in the format "{arg:value}" or "{arg1:value1,arg2:value2,...}"
    to a dictionary with keys and values.
    Note: strings should not contain ' or ".

    :param arg_str: A string in the format "{arg:value}" or
        "{arg1:value1, arg2:value2,...}".
    :return arg_dict: dictionary with keys and values corresponding to the input string.
    """
    arg_dict = {}
    # Use regular expression to extract argument pairs
    arg_pairs = re.findall(r"(\w+?):((\d+\.\d+)|(\d+)|([-\w.\[\],()]+))", arg_str)
    # Loop over argument pairs and add to dictionary
    for pair in arg_pairs:
        arg_name, arg_value = pair[0], pair[1]
        # If the value is a tuple, convert it to a tuple
        if "(" in arg_value and ")" in arg_value:
            arg_value = ast.literal_eval(arg_value)
        # If the value is a list, convert it to a list
        elif "[" in arg_value and "]" in arg_value:
            arg_value = ast.literal_eval(arg_value)
        else:
            # If the value is an integer, convert it to an integer
            try:
                arg_value = int(arg_value)
            # If the value is a float, convert it to a float
            except ValueError:
                try:
                    arg_value = float(arg_value)
                except ValueError:
                    # If the value is a string, strip the quotes
                    arg_value = arg_value.strip("'\"")
        arg_dict[arg_name] = arg_value
    return arg_dict


def parse_sequence(s):
    """
    Parses a string to a list/tuple for argparse. Can be used as type for argparse.

    :param s: String to parse.
    :return: tuple or list.
    :raises argparse.ArgumentTypeError: If the string cannot be parsed to a tuple.
    """
    try:
        # Using ast.literal_eval to safely evaluate the string as a Python literal
        return ast.literal_eval(s)
    except (ValueError, SyntaxError):
        raise argparse.ArgumentTypeError(f"Invalid list value: {s}")


def parse_list(s):
    """
    Parses a string to a list for argparse. Can be used as type for argparse.

    :param s: String to parse.
    :return: list.
    """
    return list(parse_sequence(s))


def is_between(x, min, max):
    """
    Checks if x is between min and max.

    :param x: Value to check.
    :param min: Minimum value.
    :param max: Maximum value.
    :return: True if x is between min and max, False otherwise.
    """

    return (x > min) & (x < max)


def run_imap_multiprocessing(func, argument_list, num_processes, verb=True):
    """
    Runs a function with a list of arguments in parallel using multiprocessing.

    :param func: Function to run.
    :param argument_list: List of arguments to run the function with.
    :param num_processes: Number of processes to use.
    :param verb: If True, show progress bar.
    :return: List of results from the function.
    """

    pool = Pool(processes=num_processes)

    result_list = []
    if verb:
        for result in LOGGER.progressbar(
            pool.imap(func=func, iterable=argument_list),
            total=len(argument_list),
            at_level="info",
        ):
            result_list.append(result)
    else:
        for result in pool.imap(func=func, iterable=argument_list):
            result_list.append(result)

    return result_list



import numpy as np

from cosmic_toolbox.utils import arg_str_to_dict, is_between, parse_list, parse_sequence


def test_arg_str_to_dict():
    arg_str = "{arg1:10, arg2:3.14, arg3:value}"
    expected_result = {"arg1": 10, "arg2": 3.14, "arg3": "value"}
    result = arg_str_to_dict(arg_str)
    assert result == expected_result, f"Expected {expected_result}, but got {result}"

    arg_str = "{arg:value}"
    expected_result = {"arg": "value"}
    result = arg_str_to_dict(arg_str)
    assert result == expected_result, f"Expected {expected_result}, but got {result}"

    arg_str = "{arg1:1,arg2:2,arg3:3}"
    expected_result = {"arg1": 1, "arg2": 2, "arg3": 3}
    result = arg_str_to_dict(arg_str)
    assert result == expected_result, f"Expected {expected_result}, but got {result}"

    arg_str = "{batch_size:10000,epochs:1000,hidden_units:(512,512,512)}"
    expected_result = {
        "batch_size": 10000,
        "epochs": 1000,
        "hidden_units": (512, 512, 512),
    }
    result = arg_str_to_dict(arg_str)
    assert result == expected_result, f"Expected {expected_result}, but got {result}"


def test_parse_sequence():
    s = "(1,2,3)"
    assert parse_sequence(s) == (1, 2, 3)
    s = "(1,2,3,)"
    assert parse_sequence(s) == (1, 2, 3)
    s = "[1,2,3,4]"
    assert parse_sequence(s) == [1, 2, 3, 4]


def test_parse_list():
    s = "(1,2,3)"
    assert parse_list(s) == [1, 2, 3]
    s = "(1,2,3,)"
    assert parse_list(s) == [1, 2, 3]
    s = "[1,2,3,4]"
    assert parse_list(s) == [1, 2, 3, 4]


def test_is_between():
    # Test case 1: x is between min and max
    assert is_between(5, 2, 10)

    # Test case 2: x is equal to min
    assert not is_between(2, 2, 10)

    # Test case 3: x is equal to max
    assert not is_between(10, 2, 10)

    # Test case 4: x is smaller than min
    assert not is_between(1, 2, 10)

    # Test case 5: x is greater than max
    assert not is_between(15, 2, 10)

    # Test case 6: Test with negative values
    assert is_between(-5, -10, -2)

    # Test case 7: Test with floating-point values
    assert is_between(3.5, 2.0, 5.0)

    # Test case 8: Test with min and max swapped
    assert not is_between(7, 10, 2)

    # Test case 9: Test with equal min and max
    assert not is_between(5, 5, 5)

    # Test case 10: Test with numpy array input
    np_array = np.arange(10)
    expected_result = np.array(
        [False, False, False, True, True, True, False, False, False, False]
    )
    assert np.array_equal(is_between(np_array, 2, 6), expected_result)



# Copyright (C) 2017 ETH Zurich, Cosmology Research Group

"""
Created on Aug 23, 2017
@author: Joerg Herbel
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import distutils.dir_util
import os
import random
import shlex
import shutil
import subprocess
import time

from cosmic_toolbox import file_utils, logger

LOGGER = logger.get_logger(__file__)


SEMAPHORE_DIRECTORY = os.path.expanduser("~/copy_guardian_semaphores")


if not os.path.isdir(SEMAPHORE_DIRECTORY):
    try:
        os.mkdir(SEMAPHORE_DIRECTORY)
    except OSError:
        LOGGER.warning(
            "Semaphore directory does not exist, but it could not " "be created either!"
        )


class CopyGuardian(object):
    def __init__(
        self,
        n_max_connect,
        n_max_attempts_remote,
        time_between_attempts,
        use_copyfile=False,
    ):
        self.n_max_connect = n_max_connect
        self.n_max_attempts_remote = n_max_attempts_remote
        self.time_between_attempts = time_between_attempts
        self.use_copyfile = use_copyfile

    def __call__(self, sources, destination):
        # Ensure correct type
        if str(sources) == sources:
            sources = [sources]

        if str(destination) != destination:
            raise ValueError(
                "Destination {} not supported. Multiple destinations for "
                "multiple sources not implemented.".format(destination)
            )

        # Ensure that rsync and local copy behave equally
        for i, source in enumerate(sources):
            if os.path.isdir(source) and not source.endswith("/"):
                sources[i] += "/"

        if destination.endswith("/"):
            destination = destination[:-1]

        # Check if destination is remote
        if file_utils.is_remote(destination):
            # Check for remote sources
            for source in sources:
                if file_utils.is_remote(source):
                    raise IOError(
                        "Cannot copy remote source {} to remote "
                        "destination {}".format(source, destination)
                    )

            self._copy_remote(sources, destination)

        else:
            # Split into local and remote sources
            sources_remote = []

            for source in sources:
                # Remote source
                if file_utils.is_remote(source):
                    sources_remote.append(source)

                # Local source
                else:
                    self._copy_local(source, destination)

            # Now handle remaining (remote) tasks
            if len(sources_remote) > 0:
                self._copy_remote(sources_remote, destination)

    def _copy_local(self, source, destination):
        LOGGER.info("Copying locally: {} -> {}".format(source, destination))

        if os.path.isdir(source):
            distutils.dir_util.copy_tree(source, destination)

        elif os.path.isdir(destination) or not self.use_copyfile:
            shutil.copy(source, destination)

        else:
            shutil.copyfile(source, destination)

    def _copy_remote(self, sources, destination):
        n_attempts = 0

        sources_split = self._split_sources_by_host(sources)
        print(sources_split)

        while n_attempts < self.n_max_attempts_remote:
            self._wait_for_allowance()
            path_semaphore = self._create_semaphore()

            copied = True

            for srcs in sources_split:
                for src in srcs:
                    copied &= self._call_rsync([src], destination)
                # copied &= self._call_rsync(srcs, destination)

            os.remove(path_semaphore)

            if copied:
                break

            else:
                n_attempts += 1
                time.sleep(self.time_between_attempts)
                if n_attempts * self.time_between_attempts > (5 * 60):
                    LOGGER.warning(
                        "waiting for free semaphore for long time, "
                        "n_attempts={}, time={}s".format(
                            n_attempts, n_attempts * self.time_between_attempts
                        )
                    )

        else:
            raise IOError(
                "Failed to rsync {} -> {} ".format(", ".join(sources), destination)
            )

    def _wait_for_allowance(self):
        while True:
            time.sleep(1 + random.random())

            file_list = os.listdir(SEMAPHORE_DIRECTORY)
            file_list = list(
                filter(lambda filename: not filename.startswith("."), file_list)
            )

            if len(file_list) < self.n_max_connect:
                return

    def _create_semaphore(self):
        filename = "{}_{}".format(os.getpid(), datetime.datetime.now()).replace(" ", "")
        filepath = os.path.join(SEMAPHORE_DIRECTORY, filename)
        open(filepath, "w").close()
        return filepath

    def _call_rsync(self, sources, destination):
        LOGGER.info("Rsyncing: {} -> {}".format(", ".join(sources), destination))

        cmd = "rsync -av {} {}".format(" ".join(sources), destination)
        print(cmd)

        try:
            subprocess.check_call(shlex.split(cmd))
            return True

        except subprocess.CalledProcessError:
            return False

    def _split_sources_by_host(self, sources):
        host_dict = {}

        for s in sources:
            host = s.split(":/")[0]

            if host in host_dict:
                host_dict[host].append(s)

            else:
                host_dict[host] = [s]

        return host_dict.values()



# Copyright (C) 2017 ETH Zurich
# Cosmology Research Group
# Author: Joerg Herbel


import distutils.dir_util
import os
import pickle
import shlex
import shutil
import stat
import subprocess

import copy_guardian
import h5py
import yaml

from cosmic_toolbox.logger import get_logger

DEFAULT_ROOT_PATH = ""
LOGGER = get_logger(__file__)


def robust_remove(path):
    """
    Remove a file or directory.

    :param path: Path to the file or directory.
    """
    if is_remote(path):
        LOGGER.info("Removing remote directory {}".format(path))
        host, path = path.split(":")
        cmd = 'ssh {} "rm -rf {}"'.format(host, path)
        subprocess.call(shlex.split(cmd))

    else:
        if os.path.isfile(path):
            os.remove(path)

        elif os.path.isdir(path):
            shutil.rmtree(path)

        else:
            LOGGER.warning(f"Cannot remove {path} because it does not exist")


def write_to_pickle(filepath, obj, compression="none"):
    """
    Write an object to a pickle file.

    :param filepath: Path to the pickle file.
    :param obj: Object to write.
    :param compression: Compression method to use. Can be "none", "lzf" or "bz2".
    """
    if compression.lower() == "none":
        with open(filepath, "wb") as f:
            pickle.dump(obj, f)
    elif compression.lower() == "lzf":
        import lzf

        with lzf.open(filepath, "wb") as f:
            pickle.dump(obj, f)
    elif compression.lower() == "bz2":
        import bz2

        with bz2.open(filepath, "wb") as f:
            pickle.dump(obj, f)

    else:
        raise Exception(f"uknown compression {compression} [none, lzf, bz2]")


def read_from_pickle(filepath, compression="none"):
    """
    Read an object from a pickle file.

    :param filepath: Path to the pickle file.
    :param compression: Compression method to use. Can be "none", "lzf" or "bz2".
    :return: Object read from the pickle file.
    """
    if compression.lower() == "none":
        with open(filepath, "rb") as f:
            obj = pickle.load(f)
    elif compression.lower() == "lzf":
        import lzf

        with lzf.open(filepath, "rb") as f:
            obj = pickle.load(f)
    elif compression.lower() == "bz2":
        import bz2

        with bz2.open(filepath, "rb") as f:
            obj = pickle.load(f)

    else:
        raise Exception(f"uknown compression {compression} [none, lzf, bz2]")

    return obj


def write_to_hdf(filepath, obj, name="data", **kwargs):
    """
    Write an object to an hdf5 file.

    :param filepath: Path to the hdf5 file.
    :param obj: Object to write.
    :param name: Name of the dataset.
    :param kwargs: Additional arguments passed to h5py.File.create_dataset.
    """

    with h5py.File(filepath, "w") as f:
        f.create_dataset(name, data=obj, **kwargs)


def read_from_hdf(filepath, name="data"):
    """
    Read an object from an hdf5 file.

    :param filepath: Path to the hdf5 file.
    :param name: Name of the dataset.
    :return: Object read from the hdf5 file.
    """

    with h5py.File(filepath, "r") as f:
        obj = f[name][:]

    return obj


def load_from_hdf5(file_name, hdf5_keys, hdf5_path="", root_path=DEFAULT_ROOT_PATH):
    """
    Load data stored in a HDF5-file.
    :param file_name: Name of the file.
    :param hdf5_keys: Keys of arrays to be loaded.
    :param hdf5_path: Path within HDF5-file appended to all keys.
    :param root_path: Relative or absolute root path.
    :return: Loaded arrays.
    """

    if str(hdf5_keys) == hdf5_keys:
        hdf5_keys = [hdf5_keys]
        return_null_entry = True
    else:
        return_null_entry = False

    hdf5_keys = [hdf5_path + hdf5_key for hdf5_key in hdf5_keys]

    path = get_abs_path(file_name, root_path=root_path)

    with h5py.File(path, mode="r") as hdf5_file:
        hdf5_data = [hdf5_file[hdf5_key][...] for hdf5_key in hdf5_keys]

    if return_null_entry:
        hdf5_data = hdf5_data[0]

    return hdf5_data


def get_abs_path(path):
    if "@" in path and ":/" in path:
        abs_path = path

    elif os.path.isabs(path):
        abs_path = path

    else:
        if "SUBMIT_DIR" in os.environ:
            parent = os.environ["SUBMIT_DIR"]
        else:
            parent = os.getcwd()

        abs_path = os.path.join(parent, path)

    return abs_path


def robust_makedirs(path):
    if is_remote(path):
        LOGGER.info("Creating remote directory {}".format(path))
        host, path = path.split(":")
        cmd = 'ssh {} "mkdir -p {}"'.format(host, path)
        subprocess.call(shlex.split(cmd))

    elif not os.path.isdir(path):
        os.makedirs(path)
        LOGGER.info("Created directory {}".format(path))


def robust_copy(
    src,
    dst,
    n_max_connect=50,
    method="CopyGuardian",
    folder_with_many_files=False,
    **kwargs,
):
    """
    Copy files/directories using the specified method.

    :param src: Source file/directory.
    :param dst: Destination file/directory.
    :param n_max_connect: Maximum number of simultaneous connections.
    :param method: Method to use for copying. Can be "CopyGuardian" or "system_cp".
    :param folder_with_many_files: If True, the source is a folder with many files (only for CopyGuardian).
    :param kwargs: Additional arguments passed to the copy method.
    """
    src = _ensure_list(src)
    if str(dst) != dst:
        raise ValueError(
            "Destination {} not supported. Multiple destinations for "
            "multiple sources not implemented.".format(dst)
        )

    # In case of a remote destination, rsync will create the directory itself
    if not is_remote(dst):
        robust_makedirs(os.path.dirname(dst))

    if method == "CopyGuardian":
        copy_with_copy_guardian(
            src,
            dst,
            n_max_connect=n_max_connect,
            folder_with_many_files=folder_with_many_files,
        )

    elif method == "system_cp":
        system_copy(sources=src, dest=dst, **kwargs)

    else:
        raise Exception(f"Unknown copy method {method}")


def _ensure_list(obj):
    if not isinstance(obj, list):
        return [obj]
    return obj


def copy_with_copy_guardian(
    sources,
    destination,
    n_max_connect=10,
    timeout=1000,
    folder_with_many_files=False,
):
    """
    Copy files/directories using the CopyGuardian.

    :param sources: List of source files/directories.
    :param destination: Destination directory.
    :param n_max_connect: Maximum number of simultaneous connections.
    :param timeout: time in seconds to wait for a connection to become available
    :param folder_with_many_files: If True, the source is a folder with many files
    """
    with copy_guardian.BoundedSemaphore(n_max_connect, timeout=timeout):
        for src in sources:
            LOGGER.info("Copying locally: {} -> {}".format(src, destination))
            if os.path.isdir(src):
                if folder_with_many_files:
                    LOGGER.debug("Copying folder with many files")
                    # robust_makedirs(destination)
                    copy_guardian.copy_utils.copy_local_folder(src, destination)
                else:
                    distutils.dir_util.copy_tree(src, destination)
            elif os.path.isdir(destination):
                shutil.copy(src, destination)
            else:
                shutil.copyfile(src, destination)


def system_copy(sources, dest, args_str_cp=""):
    cmd = "cp -r " + args_str_cp
    for f in sources:
        cmd += f" {f} "
    cmd += " {}".format(dest)
    LOGGER.debug("Copying {} files to {}".format(len(sources), dest))
    LOGGER.debug(cmd)
    os.system(cmd)


def is_remote(path):
    return "@" in path and ":/" in path


def read_yaml(filename):
    with open(filename) as f:
        file = yaml.load(f, Loader=yaml.Loader)
    return file


def ensure_permissions(path, verb=False):
    val = stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
    os.chmod(path, val)
    if verb:
        LOGGER.debug("Changed permissions for {} to {}".format(path, oct(val)))



import logging
import os

from cosmic_toolbox.logger import get_logger, set_all_loggers_level, set_logger_level


def test_get_logger():
    logger = get_logger("test.py")
    assert isinstance(logger, logging.Logger)


def test_set_logger_level():
    logger = logging.getLogger("test_logger")
    set_logger_level(logger, "info")
    assert logger.level == logging.INFO


def test_set_all_loggers_level():
    set_all_loggers_level("debug")
    assert os.environ["PYTHON_LOGGER_LEVEL"] == "debug"



# Copyright (C) 2023 ETH Zurich
# Institute for Particle Physics and Astrophysics


def test_import():
    import cosmic_toolbox
    import cosmic_toolbox.arraytools
    import cosmic_toolbox.colors
    import cosmic_toolbox.copy_guardian
    import cosmic_toolbox.file_utils
    import cosmic_toolbox.logger
    import cosmic_toolbox.MultiInterp
    import cosmic_toolbox.NearestWeightedNDInterpolator
    import cosmic_toolbox.TransformedGaussianMixture

    print(cosmic_toolbox.__version__)
    assert True


def test_dummy():
    assert True



"""
Convenience interface to N-D interpolation
.. versionadded:: 0.9
author: Tomasz Kacprzak
"""

import numpy as np
from scipy.interpolate.interpnd import NDInterpolatorBase
from sklearn.neighbors import BallTree
from sklearn.preprocessing import MinMaxScaler


class NearestWeightedNDInterpolator(NDInterpolatorBase):
    def __init__(self, x, y, k=None, tree_options={}):
        self.x = x
        self.y = y
        self.ndim = self.x.shape[1]
        if k is None:
            k = self.ndim + 1  # number of vertices of ndim dimensional simplex
        self.k = k
        self.scaler = MinMaxScaler()
        self.x = self.scaler.fit_transform(self.x)
        self.tree = BallTree(self.x, **tree_options)

        # from sklearn import neighbors
        # self.knn = neighbors.KNeighborsRegressor(n_neighbors=k,
        #                                          weights='distance')
        # self.knn.fit(self.x , self.y)

    def __call__(self, xi):
        assert len(xi.shape) == 2
        assert self.ndim == xi.shape[1]

        xi = self.scaler.transform(xi)
        # vi = self.knn.predict(xi)

        dist, i = self.tree.query(xi, self.k)
        vi = self.y[i].reshape((xi.shape[0], self.k))
        if self.k > 1:
            weight = 1.0 / dist
            weight[~np.isfinite(weight)] = 0
            weight = weight.reshape((xi.shape[0], self.k))
            vi = np.average(vi, weights=weight, axis=1)

        return vi


# class NearestWeightedNDInterpolator(NDInterpolatorBase):
#     """
#     NearestWeightedNDInterpolator(x, y)
#     NN interpolation in N dimensions with weighted interpolation.
#     Uses BallTree instead of cKDTree
#     .. versionadded:: 0.9
#     Methods
#     -------
#     __call__
#     Parameters
#     ----------
#     x : (Npoints, Ndims) ndarray of floats
#         Data point coordinates.
#     y : (Npoints,) ndarray of float or complex
#         Data values.
#     rescale : boolean, optional
#         Rescale points to unit cube before performing interpolation.
#         This is useful if some of the input dimensions have
#         incommensurable units and differ by many orders of magnitude.
#         .. versionadded:: 0.14.0
#     tree_options : dict, optional
#         Options passed to the underlying ``cKDTree``.
#         .. versionadded:: 0.17.0
#     Notes
#     -----
#     Uses ``scipy.spatial.cKDTree``
#     Examples
#     --------
#     We can interpolate values on a 2D plane:
#     >>> from scipy.interpolate import NearestNDInterpolator
#     >>> import matplotlib.pyplot as plt
#     >>> np.random.seed(0)
#     >>> x = np.random.random(10) - 0.5
#     >>> y = np.random.random(10) - 0.5
#     >>> z = np.hypot(x, y)
#     >>> X = np.linspace(min(x), max(x))
#     >>> Y = np.linspace(min(y), max(y))
#     >>> X, Y = np.meshgrid(X, Y)  # 2D grid for interpolation
#     >>> interp = NearestNDInterpolator(list(zip(x, y)), z)
#     >>> Z = interp(X, Y)
#     >>> plt.pcolormesh(X, Y, Z, shading='auto')
#     >>> plt.plot(x, y, "ok", label="input point")
#     >>> plt.legend()
#     >>> plt.colorbar()
#     >>> plt.axis("equal")
#     >>> plt.show()
#     See also
#     --------
#     griddata :
#         Interpolate unstructured D-D data.
#     LinearNDInterpolator :
#         Piecewise linear interpolant in N dimensions.
#     CloughTocher2DInterpolator :
#         Piecewise cubic, C1 smooth, curvature-minimizing interpolant in 2D.
#     """

#     def __init__(self, x, y, k=None, tree_options={}):
#         NDInterpolatorBase.__init__(self, x, y, rescale=True,
#                                     need_contiguous=False,
#                                     need_values=False)
#         ndim = self.points.shape[1]
#         if k==None:
#             k = ndim + 1 # number of vertices of ndim dimensional simplex
#         self.k = k
#         self.scaler = MinMaxScaler(copy=False)
#         self.scaler.fit_transform(self.points)
#         self.tree = BallTree(self.points, **tree_options)
#         self.values = np.asarray(y)

#     def __call__(self, *args):
#         """
#         Evaluate interpolator at given points.
#         Parameters
#         ----------
#         x1, x2, ... xn: array-like of float
#             Points where to interpolate data at.
#             x1, x2, ... xn can be array-like of float with broadcastable
#             shape or x1 can be array-like of float with shape ``(..., ndim)``
#         """

#         ndim = self.points.shape[1]
#         xi = _ndim_coords_from_arrays(args, ndim=ndim)
#         xi = self._check_call_shape(xi)
#         # xi = self._scale_x(xi)
#         orig_shape = xi.shape

#         if len(xi.shape) > 2:
#             xi = xi.reshape(-1, xi.shape[-1])
#         self.scaler.transform(xi)

#         dist, i = self.tree.query(xi, self.k)
#         weight = 1./dist
#         print(weight.shape)
#         weight[~np.isfinite(weight)] = 0
#         import ipdb; ipdb.set_trace()
#         try:
#             vi = np.average(self.values[i].squeeze(), weights=weight, axis=1)
#         except Exception as err:
#             print(err)
#             import ipdb; ipdb.set_trace()

#         return np.reshape(vi, orig_shape[:-1])



# Copyright (C) 2017 ETH Zurich
# Cosmology Research Group
# Author: Joerg Herbel, Silvan Fischbacher


import logging
import os
import sys

logging_levels = {
    "critical": logging.CRITICAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "info": logging.INFO,
    "debug": logging.DEBUG,
}


def get_logger(filepath, logging_level=None):
    """
    Get logger, if logging_level is unspecified, then try using the
    environment variable PYTHON_LOGGER_LEVEL.
    Defaults to info.
    :param filepath: name of the file that is calling the logger,
        used to give it a name.
    :return: logger object
    """

    if logging_level is None:
        if "PYTHON_LOGGER_LEVEL" in os.environ:
            logging_level = os.environ["PYTHON_LOGGER_LEVEL"]
        else:
            logging_level = "info"

    logger = logging.getLogger(os.path.basename(filepath)[:10])

    if len(logger.handlers) == 0:
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(ColorFormatter())
        logger.addHandler(stream_handler)
        logger.propagate = False
        set_logger_level(logger, logging_level)

    logger.progressbar = Progressbar(logger)

    return logger


class ColorFormatter(logging.Formatter):
    RED = "\033[91m"
    VIOLET = "\033[95m"
    YELLOW = "\033[93m"
    ORANGE = "\033[33m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    reset = "\033[0m"
    format = "%(asctime)s %(name)10s %(levelname).3s   %(message)s "

    FORMATS = {
        logging.DEBUG: VIOLET + format + reset,
        logging.INFO: format,
        logging.WARNING: BOLD + ORANGE + format + reset,
        logging.ERROR: BOLD + RED + format + reset,
        logging.CRITICAL: BOLD + RED + format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt="%y-%m-%d %H:%M:%S")
        return formatter.format(record)


class Progressbar:
    def __init__(self, logger=None):
        self.logger = logger

    def __call__(self, collection, at_level="info", **kw):
        kw.setdefault("bar_format", "{percentage:3.0f}%|{bar:28}|   {r_bar:<40} {desc}")
        kw.setdefault("disable", self.logger.level != logging_levels[at_level])
        kw.setdefault("colour", "blue")
        kw.setdefault("mininterval", 1)
        kw.setdefault("file", sys.stdout)
        from tqdm import tqdm

        return tqdm(collection, **kw)


def set_logger_level(logger, level):
    logger.setLevel(logging_levels[level])


def set_all_loggers_level(level):
    os.environ["PYTHON_LOGGER_LEVEL"] = level

    loggerDict = logging.root.manager.loggerDict
    for key in loggerDict:
        try:
            set_logger_level(logger=loggerDict[key], level=level)
        except Exception as err:
            LOGGER.debug(err)


LOGGER = get_logger(__file__)



import matplotlib.pyplot as plt
from cycler import cycler


def get_colors(cycle="silvan"):
    if cycle == "silvan":
        return {
            "b": "#0063B9",
            "r": "#BA1A13",
            "g": "#34BA09",
            "orange": "#ED6018",
            "violet": "#6E13BA",
            "brown": "#6E2C0B",
            "gray": "#808080",
            "pink": "#BA13A0",
            "olive": "#B0BA13",
            "bluegreen": "#09BAAF",
            "lightblue": "#1994FF",
            "lightred": "#FF3D33",
            "lightgreen": "#5CFF26",
            "lightorange": "#FFA547",
            "lightviolet": "#A333FF",
            "lightbrown": "#FFA541",
            "lightgray": "#BEBEBE",
            "lightpink": "#FF33E0",
            "lightolive": "#F1FF33",
            "lightbluegreen": "#33FFE7",
        }
    else:
        return cycle


def set_cycle(cycle="silvan"):
    colors = get_colors(cycle=cycle)
    if isinstance(colors, dict):
        col = colors.values()
    else:
        col = colors
    plt.rc("axes", prop_cycle=cycler(color=col))



import os

import h5py
import numpy as np
import pandas as pd
import six

from cosmic_toolbox import file_utils, logger

LOGGER = logger.get_logger(__file__)


def view_fields(rec, names):
    """
    `rec` must be a numpy structured array.
    `names` is the collection of field names to keep.

    Returns a view of the array `a` (not a copy).
    """
    dt = rec.dtype
    formats = [dt.fields[name][0] for name in names]
    offsets = [dt.fields[name][1] for name in names]
    itemsize = rec.dtype.itemsize
    newdt = np.dtype(
        dict(names=names, formats=formats, offsets=offsets, itemsize=itemsize)
    )
    rec_view = rec.view(newdt)
    return rec_view


def delete_cols(rec, col_names):
    """
    Delete columns from a numpy recarray.

    :param rec: numpy recarray
    :param col_names: list of names of the columns to delete
    :return: numpy recarray
    """
    col_names_all = list(rec.dtype.names)

    for col_name in col_names:
        if col_name in col_names_all:
            col_names_all.remove(col_name)

    if not col_names_all:
        # If no columns are left, return an empty recarray with an empty dtype
        return np.array([], dtype=[])

    return rec[col_names_all].copy()


def delete_columns(rec, col_names):
    """
    Delete columns from a numpy recarray. (alias for delete_cols for backwards compatibility)

    :param rec: numpy recarray
    :param col_names: list of names of the columns to delete
    """
    return delete_cols(rec, col_names)


def add_cols(rec, names, shapes=None, data=0, dtype=None):
    """
    Add columns to a numpy recarray. By default, the new columns are filled with zeros.
    If `data` is a numpy array, it is used to fill the new columns. If each column should
    should be filled with different data, `data` should be a list of numpy arrays or an
    array of shape (n_cols, n_rows).

    :param rec: numpy recarray
    :param names: list of names for the columns
    :param shapes: list of shapes for the columns
    :param data: data to fill the columns with
    :param dtype: dtype of the columns
    :return: numpy recarray
    """

    # check if new data should be sliced
    slice_data = isinstance(data, np.ndarray) and data.ndim == 2

    # create new recarray
    names = [str(name) for name in names]
    extra_dtype = get_dtype(names, shapes=shapes, main=dtype)
    newdtype = np.dtype(rec.dtype.descr + extra_dtype.descr)
    newrec = np.empty(rec.shape, dtype=newdtype)

    # add data to new recarray
    for field in rec.dtype.fields:
        newrec[field] = rec[field]
    for ni, name in enumerate(extra_dtype.fields):
        if slice_data:
            newrec[name] = data[ni]
        else:
            newrec[name] = np.array(data).astype(dtype)

    return newrec


def ensure_cols(rec, names, shapes=None, data=0):
    # find columns to add
    names = [str(name) for name in names]
    new_names = []
    for name in names:
        if get_dtype([name]).names[0] not in rec.dtype.names:
            new_names.append(name)

    # exit if no new cols
    if len(new_names) == 0:
        return rec

    # add new columns
    newrec = add_cols(rec, new_names, shapes=shapes, data=data)

    return newrec


def arr2rec(arr, names):
    """
    Convert a numpy array to a numpy structured array.

    :param arr: numpy array
    :param names: list of names for the columns
    :return: numpy structured array
    """
    arr = np.atleast_2d(arr)
    n_arr, n_params = arr.shape
    dtype = dict(formats=[], names=names)
    for i in range(n_params):
        dtype["formats"].append(arr[:, i].dtype)
    rec = np.zeros(n_arr, dtype=dtype)
    for i, name in enumerate(names):
        rec[name] = arr[:, i]
    return rec


def rec2arr(rec, return_names=False):
    """
    Convert a numpy structured array to a numpy array.

    :param rec: numpy structured array
    :param return_names: if True, also return the names of the columns
    :return: numpy array

    Example
    -------
    >>> rec = np.array([(1, 4), (2, 4), (3, 4)],
            dtype=[('a', '<i8'), ('b', '<i8')])
    >>> arr = rec2arr(rec)
    >>> arr
    array([[1, 4], [2, 4], [3, 4]])
    >>> arr, names = rec2arr(rec, return_names=True)
    >>> arr
    array([[1, 4], [2, 4], [3, 4]])
    >>> names
    ['a', 'b']
    """
    arr = np.array([rec[par] for par in rec.dtype.names]).T
    if return_names:
        return arr, rec.dtype.names
    else:
        return arr


def dict2rec(d):
    """
    Convert a dictionary of arrays/lists to a numpy structured array.

    :param d: Dictionary with arrays/lists as values
    :return: numpy structured array
    """

    # Convert lists/values to numpy arrays
    d = {
        k: np.array(v) if hasattr(v, "__iter__") and not isinstance(v, str) else v
        for k, v in d.items()
    }

    # Get first dimension size from array values
    array_sizes = [v.shape[0] for v in d.values() if hasattr(v, "shape")]
    if not array_sizes:
        raise ValueError("At least one array value required")
    size = array_sizes[0]

    # Create dtype with proper shapes
    dtype_list = []
    for key, val in d.items():
        if hasattr(val, "dtype"):
            if val.ndim > 1:
                dtype_list.append((key, val.dtype, val.shape[1:]))
            else:
                dtype_list.append((key, val.dtype))
        else:
            # For scalar values
            dtype_list.append((key, np.array(val).dtype))

    # Create empty structured array
    rec = np.empty(size, dtype=dtype_list)

    # Fill array
    for key, val in d.items():
        if hasattr(val, "shape"):
            rec[key] = val
        else:
            rec[key] = np.full(size, val)

    return rec


def rec2dict(rec):
    """
    Convert a numpy structured array to a dictionary.

    :param rec: numpy structured array
    :return: dictionary

    Example
    -------
    >>> rec = np.array([(1, 4), (2, 4), (3, 4)],
            dtype=[('a', '<i8'), ('b', '<i8')])
    >>> d = rec2dict(rec)
    >>> d
    {'a': array([1, 2, 3]), 'b': array([4, 4, 4])}
    """
    keys = rec.dtype.names
    d = {}
    for key in keys:
        d[key] = rec[key]
    return d


def dict2class(d):
    """
    Convert a dictionary to a class.

    :param d: dictionary
    :return: class

    Example
    -------
    >>> d = {'a': [1, 2, 3], 'b': 4}
    >>> c = dict2class(d)
    >>> c.a
    [1, 2, 3]
    >>> c.b
    4
    """

    class C:
        pass

    c = C()
    for key in d.keys():
        setattr(c, key, d[key])
    return c


def rec2class(rec):
    """
    Convert a numpy structured array to a class.

    :param rec: numpy structured array
    :return: class
    """
    return dict2class(rec2dict(rec))


def class2dict(c):
    """
    Convert a class to a dictionary.

    :param c: class
    :return: dictionary
    """

    return vars(c)


def class2rec(c):
    """
    Convert a class to a numpy structured array.

    :param c: class
    :return: numpy structured array
    """
    return dict2rec(class2dict(c))


def pd2rec(df):
    """
    Convert a pandas dataframe to a numpy structured array.

    :param df: pandas dataframe
    :return: numpy structured array
    """
    return df.to_records(index=False)


def rec2pd(rec):
    data_dict = {}
    for name in rec.dtype.names:
        if rec.dtype[name].shape:
            for i in range(rec.dtype[name].shape[0]):
                data_dict[f"{name}_{i}"] = rec[name][:, i]
        else:
            # Otherwise, just add it as a regular column
            data_dict[name] = rec[name]
    df = pd.DataFrame(data_dict)
    return df


def get_nan_mask(rec):
    """
    Get a mask for rows with NaNs in a numpy structured array.

    :param rec: numpy structured array
    :return: numpy structured array
    """
    nan_mask = np.zeros(len(rec), dtype=bool)
    for name in rec.dtype.names:
        nan_mask |= np.isnan(rec[name])
    return nan_mask


def remove_nans(rec):
    """
    Remove rows with NaNs from a numpy structured array.

    :param rec: numpy structured array
    :return: numpy structured array
    """
    nan_mask = get_nan_mask(rec)
    return rec[~nan_mask]


def get_inf_mask(rec):
    """
    Get a mask for rows with infs in a numpy structured array.

    :param rec: numpy structured array
    :return: numpy structured array
    """
    inf_mask = np.zeros(len(rec), dtype=bool)
    for name in rec.dtype.names:
        inf_mask |= np.isinf(rec[name])
    return inf_mask


def remove_infs(rec):
    """
    Remove rows with infs from a numpy structured array.

    :param rec: numpy structured array
    :return: numpy structured array
    """
    inf_mask = get_inf_mask(rec)
    return rec[~inf_mask]


def get_finite_mask(rec):
    """
    Get a mask for finite rows (i.e., rows without NaNs or infs) in a numpy structured array.

    :param rec: numpy structured array
    :return: numpy structured array
    """
    inf_mask = get_inf_mask(rec)
    nan_mask = get_nan_mask(rec)
    return ~(inf_mask | nan_mask)


def select_finite(rec):
    """
    Remove rows with NaNs or infs from a numpy structured array.

    :param rec: numpy structured array
    :return: numpy structured array
    """
    return remove_infs(remove_nans(rec))


def arr_to_rec(arr, dtype):
    """
    Convert a numpy array to a numpy structured array given its dtype.

    :param arr: numpy array
    :param dtype: dtype of the structured array
    :return: numpy structured array
    """
    newrecarray = np.core.records.fromarrays(np.array(arr).transpose(), dtype=dtype)
    return newrecarray


def get_dtype_of_list(lst):
    # Convert the list to a NumPy array
    arr = np.array(lst)

    # Get the data type of the first element in the array
    dtype = arr[0].dtype
    if len(arr) == 1:
        return dtype

    # Check if the data type of every other element is the same as the first element
    all_the_same = True
    for element in arr[1:]:
        if element.dtype != dtype:
            all_the_same = False

    assert all_the_same, "Not all entries of the list have the same dtype"

    return dtype


def new_array(n_rows, columns, ints=[], float_dtype=np.float64, int_dtype=np.int64):
    n_columns = len(columns)
    formats = [None] * n_columns
    for ic in range(n_columns):
        if columns[ic] in ints:
            formats[ic] = int_dtype
        else:
            formats[ic] = float_dtype
    newrec = np.zeros(n_rows, dtype=np.dtype(zip(columns, formats)))
    return newrec


def get_dtype(columns, main="f8", shapes=None):
    if main is None:
        main = "f8"
    elif isinstance(main, type):
        main = main.__name__
    list_name = []
    list_dtype = []

    if shapes is None:
        shapes = [() for _ in columns]

    for col in columns:
        if ":" in col:
            name, dtype = col.split(":")
        else:
            name, dtype = col, main

        list_name.append(str(name))
        list_dtype.append(str(dtype))

    dtype = np.dtype(list(zip(list_name, list_dtype, shapes)))

    return dtype


def get_storing_dtypes(dtype_list):
    if six.PY2:
        dtypes = [(str(dt[0]), str(dt[1])) + dt[2:] for dt in dtype_list]

    else:
        dtypes = [(dt[0], dt[1].replace("<U", "|S")) + dt[2:] for dt in dtype_list]

    return dtypes


def get_loading_dtypes(dtype_list):
    if six.PY2:
        dtypes = [(str(dt[0]), str(dt[1])) + dt[2:] for dt in dtype_list]

    else:
        dtypes = []
        # handling for h5py>=2.10.0
        for dt in dtype_list:
            if isinstance(dt[1], tuple):
                dtypes.append((dt[0], dt[1][0].replace("|S", "<U")) + dt[2:])
            else:
                dtypes.append((dt[0], dt[1].replace("|S", "<U")) + dt[2:])

    return dtypes


def set_storing_dtypes(arr):
    if six.PY2:
        return arr

    # numpy array
    elif hasattr(arr, "dtype"):
        dtype = arr.dtype

        # ndarray
        if len(dtype) == 0:
            dtypes_store = arr.dtype.str.replace("<U", "|S")

        # recarray
        else:
            dtypes_store = get_storing_dtypes(arr.dtype.descr)

        return arr.astype(dtypes_store)

    # string
    elif isinstance(arr, str):
        return arr.encode("utf-8")

    else:
        # list or tuple
        try:
            arr_fixed = [None] * len(arr)

            for i, x in enumerate(arr):
                if isinstance(x, str):
                    arr_fixed[i] = x.encode("utf-8")
                else:
                    arr_fixed[i] = x

            return arr_fixed

        # single number
        except Exception:
            return arr


def set_loading_dtypes(arr):
    if six.PY2:
        return arr

    # numpy array
    elif hasattr(arr, "dtype"):
        dtype = arr.dtype

        # ndarray
        if len(dtype) == 0:
            dtypes_load = arr.dtype.str.replace("|S", "<U").replace("|O", "<U")

        # recarray
        else:
            dtypes_load = get_loading_dtypes(arr.dtype.descr)

        return arr.astype(dtypes_load)

    # bytes
    elif isinstance(arr, bytes):
        return arr.decode("utf-8")

    else:
        # list or tuple
        try:
            arr_fixed = [None] * len(arr)

            for i, x in enumerate(arr):
                if isinstance(x, bytes):
                    arr_fixed[i] = x.decode("utf-8")
                else:
                    arr_fixed[i] = x

            return arr_fixed

        # single number
        except Exception:
            return arr


def save_hdf(filename, arr, **kwargs):
    f5 = h5py.File(name=filename, mode="a")

    try:
        f5.clear()

    except Exception:
        for datasetname in f5.keys():
            del f5[datasetname]

    arr_store = set_storing_dtypes(arr)
    f5.create_dataset(name="data", data=arr_store, **kwargs)
    f5.close()
    LOGGER.info("saved {}".format(filename))


def write_to_hdf(filename, arr, name="data", compression="lzf", shuffle=True, **kwargs):
    """
    Write a recarray to a hdf file.

    :param filename: filename of the hdf file
    :param arr: numpy recarray
    :param name: name of the dataset
    :param compression: compression method
    :param shuffle: shuffle data before compression
    :param kwargs: keyword arguments for h5py.File.create_dataset
    """
    with h5py.File(filename, "w") as fh5:
        fh5.create_dataset(
            name=name, data=arr, compression=compression, shuffle=shuffle, **kwargs
        )


def load_hdf(filename, first_row=-1, last_row=-1):
    f5 = h5py.File(name=filename, mode="r")

    if (first_row > -1) & (last_row > -1) & (first_row < last_row):
        if first_row < last_row:
            data = np.array(f5["data"][first_row:last_row])

        else:
            raise Exception(
                "first_row ({}) should be smaller " "than last_row ({})".format(
                    first_row, last_row
                )
            )
    else:
        data = np.array(f5["data"])

    f5.close()

    safe_dtypes = get_loading_dtypes(data.dtype.descr)
    data = data.astype(safe_dtypes, copy=False)

    LOGGER.debug("loaded {}".format(filename))

    return data


def save_hdf_cols(filename, arr, compression=None, resizable=False):
    if compression is None:
        kwargs = {}
    else:
        if isinstance(compression, dict):
            kwargs = compression
        else:
            kwargs = {"compression": compression}

    dtypes_store = get_storing_dtypes(arr.dtype.descr)

    with h5py.File(filename, "w") as fh5:
        for dt in dtypes_store:
            col = dt[0]
            dtype = dt[1]
            extra_dims = dt[2:]

            if col in fh5:
                del fh5[col]

            if resizable:
                kwargs["maxshape"] = (None,) + extra_dims

            fh5.create_dataset(
                name=col, data=arr[col].astype(dtype, copy=False), **kwargs
            )

    LOGGER.info("saved hdf file {} with {} rows".format(filename, len(arr)))


def load_hdf_cols_from_file(
    filename,
    columns="all",
    first_row=0,
    last_row=-1,
    cols_to_add=(),
    selectors=None,
    verb=True,
):
    """
    Loads all columns of an hdf file into one recarray.

    :param filename: path to hdf file
    :param columns: list of columns to load, "all" to load all columns
    :param first_row: first row to load
    :param last_row: last row to load
    :param cols_to_add: list of columns to add to the recarray
    :param selectors: dictionary of selection masks for columns
    :param verb: if True, print information
    :return: recarray
    """
    with h5py.File(filename, mode="r") as fh5:
        # Infer columns to load
        if columns == "all":
            columns = list(fh5.keys())

        # Infer size of data to load
        if last_row is not None:
            if last_row > 0:
                size = last_row - first_row
            else:
                size = len(fh5[columns[0]]) + last_row
                last_row = size

                if size < 0 or last_row < first_row:
                    raise Exception(
                        "Combination first_row={}, last_row={} " "invalid".format(
                            first_row, last_row
                        )
                    )
        else:
            size = len(fh5[columns[0]])
            last_row = size

        # Get boolean mask in case any selectors were specified
        if selectors is not None and len(selectors) > 0:
            select = np.ones(size, dtype=np.bool)

            for cols in selectors:
                select_fun = selectors[cols]

                if isinstance(cols, str):
                    cols = (cols,)

                select &= select_fun(*[fh5[c][first_row:last_row] for c in cols])

            size = np.count_nonzero(select)

            if verb:
                LOGGER.info(
                    "masking {} / {} rows while " "loading {}".format(
                        select.size - size, select.size, filename
                    )
                )

        else:
            select = np.s_[:size]

        # Create array holding loaded data
        dtype_list = [(col, fh5[col].dtype.str, fh5[col].shape[1:]) for col in columns]
        dtypes_load = get_loading_dtypes(dtype_list)
        dtypes_load += [(col, np.float32) for col in cols_to_add]
        arr = np.empty(size, dtype=dtypes_load)

        for col in columns:
            arr[col] = fh5[col][first_row:last_row][select]

    for col in cols_to_add:
        arr[col] = 0.0

    return arr


def col_name_to_path(dirname, colname):
    return os.path.join(dirname, colname + ".h5")


def get_hdf_col_names(path):
    # file
    if os.path.isfile(path):
        with h5py.File(path, mode="r") as fh5:
            columns = list(fh5.keys())

    # directory
    columns = [
        name for name, ext in map(os.path.splitext, os.listdir(path)) if ext == ".h5"
    ]

    return columns


def load_hdf_cols_from_directory(
    dirname,
    columns="all",
    first_row=0,
    last_row=-1,
    copy_local=False,
    dirname_parent=None,
    allow_nonexisting=False,
    cols_to_add=(),
    selectors=None,
    verb=True,
    copy_editor=None,
):
    """
    Loads columns that are stored as individual files in a directory into one recarray.
    """
    if selectors is None:
        selectors = dict()

    if copy_editor is None:
        copy_editor = lambda p: p  # noqa

    # Infer columns to load
    columns_all = []
    if not allow_nonexisting or os.path.exists(dirname):
        columns_all += [
            name
            for name, ext in map(os.path.splitext, os.listdir(dirname))
            if ext == ".h5"
        ]
    if dirname_parent is not None:
        columns_all += [
            name
            for name, ext in map(os.path.splitext, os.listdir(dirname_parent))
            if ext == ".h5"
        ]
    columns_all = list(set(columns_all))

    if columns == "all":
        columns = columns_all
    else:
        columns = list(columns)

    # Get filename
    # (for all columns s.t. one can also use selectors
    # using columns that are not loaded)
    dict_filename_cols = {}
    for col in set(columns_all + columns + list(selectors.keys())):
        # get the filename from the right folder
        filename_col = col_name_to_path(dirname, col)

        if not os.path.isfile(filename_col) and dirname_parent is not None:
            filename_col = col_name_to_path(dirname_parent, col)

        dict_filename_cols[col] = filename_col

    # Infer size of data to load
    if last_row is not None:
        if last_row > 0:
            size = last_row - first_row
        else:
            with h5py.File(dict_filename_cols[columns[0]], mode="r") as fh5:
                size = len(fh5[columns[0]]) + last_row
                last_row = size

                if size < 0 or last_row < first_row:
                    raise Exception(
                        "Combination first_row={}, last_row={} invalid".format(
                            first_row, last_row
                        )
                    )
    else:
        with h5py.File(dict_filename_cols[columns[0]], mode="r") as fh5:
            size = len(fh5[columns[0]])
            last_row = size

    # Get boolean mask in case any selectors were specified
    columns_copied = {}

    if len(selectors) > 0:
        select = np.ones(size, dtype=np.bool)

        for cols in selectors:
            select_fun = selectors[cols]

            if isinstance(cols, str):
                cols = (cols,)

            data_select = [None] * len(cols)

            for ic, c in enumerate(cols):
                filename_col = dict_filename_cols[c]

                if copy_local:
                    if c not in columns_copied:
                        path_local = os.path.join(
                            os.getcwd(), os.path.basename(filename_col)
                        )
                        file_utils.robust_copy(
                            copy_editor(filename_col),
                            path_local,
                            use_copyfile=True,
                        )
                        path_load = path_local
                        columns_copied[c] = path_load
                    else:
                        path_load = columns_copied[c]

                else:
                    path_load = filename_col

                with h5py.File(path_load, mode="r") as fh5:
                    data_select[ic] = fh5[c][first_row:last_row]

            select &= select_fun(*data_select)

        size = np.count_nonzero(select)
        del data_select

        if verb:
            LOGGER.info(
                "masking {} / {} rows while loading " "{}".format(
                    select.size - size, select.size, dirname
                )
            )

    else:
        select = np.s_[:size]

    # Create array holding loaded data
    dtypes_list = [None] * len(columns)

    for i, col in enumerate(columns):
        with h5py.File(dict_filename_cols[col], mode="r") as fh5:
            dtypes_list[i] = (col, fh5[col].dtype.str, fh5[col].shape[1:])

    dtypes_load = get_loading_dtypes(dtypes_list)
    dtypes_load += [(col, np.float32) for col in cols_to_add]
    arr = np.empty(size, dtype=dtypes_load)

    # Now copy files one by one and read data
    for col in columns:
        filename_col = dict_filename_cols[col]

        if copy_local:
            if col in columns_copied:
                path_local = columns_copied[col]

            else:
                path_local = os.path.join(os.getcwd(), os.path.basename(filename_col))
                file_utils.robust_copy(
                    copy_editor(filename_col), path_local, use_copyfile=True
                )
                columns_copied[col] = path_local

            path_load = path_local

        else:
            path_load = filename_col

        with h5py.File(path_load, mode="r") as fh5:
            arr[col] = fh5[col][first_row:last_row][select]

        if copy_local:
            os.remove(path_load)
            del columns_copied[col]

    # Remove remaining columns (columns used for selecting only)
    for path_local in columns_copied.values():
        os.remove(path_local)

    # Set newly added columns to zero
    for col in cols_to_add:
        arr[col] = 0.0

    return arr


def load_hdf_cols(
    filename,
    columns="all",
    first_row=0,
    last_row=None,
    verb=True,
    copy_local=True,
    filename_parent=None,
    allow_nonexisting=False,
    cols_to_add=(),
    selectors=None,
    copy_editor=None,
):
    if last_row is not None:
        if (last_row > 0) & (first_row >= last_row):
            raise Exception(
                "first_row {} should be smaller than " "last_row {}".format(
                    first_row, last_row
                )
            )

    if os.path.isfile(filename):
        arr = load_hdf_cols_from_file(
            filename,
            columns=columns,
            first_row=first_row,
            last_row=last_row,
            cols_to_add=cols_to_add,
            selectors=selectors,
            verb=verb,
        )

    else:
        arr = load_hdf_cols_from_directory(
            filename,
            columns=columns,
            first_row=first_row,
            last_row=last_row,
            copy_local=copy_local,
            dirname_parent=filename_parent,
            allow_nonexisting=allow_nonexisting,
            cols_to_add=cols_to_add,
            selectors=selectors,
            verb=verb,
            copy_editor=copy_editor,
        )

    if verb:
        LOGGER.info("loaded {} with n_rows={}".format(filename, len(arr)))

    return arr


def append_rows_to_h5dset(dset, array):
    nr_dset, nr_array = dset.shape[0], array.shape[0]
    nr_new = nr_dset + nr_array
    dset.resize(nr_new, axis=0)
    dset[nr_dset:] = set_storing_dtypes(array)


def replace_hdf5_dataset(fobj, name, data, **kwargs):
    if name in fobj:
        del fobj[name]
    fobj.create_dataset(name=name, data=set_storing_dtypes(data), **kwargs)


def overwrite_hdf5_column(path, name, data, **kwargs):
    if os.path.isdir(path):
        path_col = col_name_to_path(path, name)

    else:
        path_col = path

    with h5py.File(path_col) as fh5:
        replace_hdf5_dataset(fh5, name, set_storing_dtypes(data), **kwargs)


def check_hdf_column(filename, column_name):
    if os.path.isfile(filename):
        with h5py.File(filename, mode="r") as fh5:
            col_found = column_name in fh5.keys()

    else:
        filename_col = col_name_to_path(filename, column_name)
        col_found = os.path.isfile(filename_col)

    return col_found


def rec_float64_to_float32(cat):
    list_new_dtype = []

    all_ok = True

    for i in range(len(cat.dtype)):
        if cat.dtype[i] == np.float64:
            list_new_dtype.append(np.float32)
            all_ok = False

        else:
            list_new_dtype.append(cat.dtype[i])

    if all_ok:
        return cat

    else:
        new_dtype = np.dtype(dict(formats=list_new_dtype, names=cat.dtype.names))
        cat_new = cat.astype(new_dtype)
        return cat_new


def save_dict_to_hdf5(filename, data_dict, kw_compress={}):
    kw_compress.setdefault("compression", "lzf")
    kw_compress.setdefault("shuffle", True)

    f = h5py.File(filename, "w")
    for grp_name in data_dict:
        grp = f.create_group(grp_name)
        for dset_name in data_dict[grp_name]:
            grp.create_dataset(
                dset_name, data=data_dict[grp_name][dset_name], **kw_compress
            )
    f.close()


def nanequal(a, b):
    return (np.isnan(b) & np.isnan(a)) | (a == b)
